#!/usr/bin/python
#-----------------------------------------------------------------------
# File    : special.py
# Contents: detect synchronous spiking activity and reduce results
# Author  : Christian Borgelt
# History : 2013.04.30 file created from detect.py
#           2014.05.27 modules placed in sibling directories
#           2015.08.12 adapted to modified pattern set reduction
#           2015.08.18 changed to function patred() from pyfim library
#-----------------------------------------------------------------------
from sys     import argv, path
from os.path import join
from time    import strftime
from random  import seed
UTILDIR  = join('..', 'util')
if UTILDIR  not in path: path.insert(0, UTILDIR)
from id2par import get_args, project, run
NEURODIR = join('..', 'neuro')
if NEURODIR not in path: path.insert(0, NEURODIR)
from psts   import genpsts, psts2tras
try:
    from fim   import fim, patred
except ImportError:
    from pyfim import fim
    FIMDIR   = join('..', 'fim')
    if FIMDIR   not in path: path.insert(0, FIMDIR)
    from patred import patred
    print('Attention: Could not import from PyFIM library!')
    print('           Using Python replacement (very slow)!')

#-----------------------------------------------------------------------

oo = float('inf')               # positive infinity as a constant
redids = ['x', 'c', 'C', 'i', 's', 'S', 'l', 'L', 't', 'T']

#-----------------------------------------------------------------------

def detect (ncnt=100, period=3, rate=20,         # neuron    parameters
            size=5, coin=5, jitter=0.003,        # assembly  parameters
            width=0.003, target='c', rfnids=[0], # detection parameters
            srng=1):                             # other     parameters
    if jitter < 0: jitter = width
    srng += 1023*size           # init. seed and get data file name
    base = 'n%d_p%g_r%g_w%g_%c' % (ncnt,period,rate,width*1000,target)
    psp  = set()                # initialize a pattern spectrum
    with open(join('data','pc_'+base+'.txt'), 'r') as inp:
        for line in inp:        # collect the pattern signatures
            z,c,n = line.split()
            psp.add((int(z),int(c)))
    border = list(range(ncnt+1)); m = 1
    for z in reversed(border):  # determine the detection border
        m = max([c for a,c in psp if a == z]+[m])
        border[z] = m+1         # (support thresholds per size)
    border[0:2] = [oo,oo]       # entirely rule out sizes 0 and 1
    base = 'n%d_p%g_r%g_j%g_w%g_%c' \
         % (ncnt,period,rate,jitter*1000,width*1000,target)
    ext  = '_%d_%d' % (size,coin)
    name = base +ext            # form the file name (parts)
    asm  = set(range(1,size+1)) # create the assembly to inject
    upcs = dict()               # and the pattern count dictionaries
    fpcs = dict()               # (unfiltered, filtered, reduced)
    rpcs = dict([(r,dict()) for r in rfnids])
    seed(srng)                  # seed random number generator
    psts = genpsts(ncnt, period, rate, jitter=jitter,
                   asms=[[asm,-coin]])
    pats = fim(psts2tras(psts, width, period),
               target=target, supp=-2, zmin=2)
    fltd = []                   # generate and mine spike trains
    for p,c in pats:            # and traverse the found patterns
        z = len(p)              # get size and support and
        s = (z,c)               # create the pattern signature
        if s in upcs: upcs[s] += 1 # count unfiltered patterns
        else:         upcs[s]  = 1 # by their signatures
        if c < border[z]: continue # collect non-excluded patterns
        fltd.append((frozenset(p),c))
        if s in fpcs: fpcs[s] += 1 # count filtered patterns
        else:         fpcs[s]  = 1 # by their signatures
    for r in rfnids:            # traverse the reduction identifiers
        pats = patred(fltd, redids[r], border, False)
        pcs  = rpcs[r]          # reduce the filtered patterns
        for p,c in pats:        # traverse the reduced patterns and
            z = len(p); s = (z,c)        # count them by signatures
            if s in pcs: pcs[s] += 1
            else:        pcs[s]  = 1
    with open(join('data','su_'+name+'.txt'), 'w') as out:
        for x in sorted([(s[0], s[1], upcs[s]) for s in upcs]):
            out.write('%g %g %g\n' % x)
    for r in rfnids:            # traverse the reduction ids and
        pcs = rpcs[r]           # write the pattern counters
        with open(join('data','sr_'+base+str(r)+ext+'.txt'),'w') as out:
            for x in sorted([(s[0], s[1], pcs[s]) for s in pcs]):
                out.write('%g %g %g\n' % x)
    print(strftime('%Y.%m.%d %H:%M:%S special ')+name+' done.')
    return (upcs, fpcs, rpcs)   # return the pattern counters

#-----------------------------------------------------------------------

if __name__ == '__main__':
    srng  = int(argv[2]) if len(argv) > 2 else 43
    grids = [ [ ( 'ncnt',   [  100 ] ),
                ( 'period', [    3 ] ),
                ( 'rate',   [   20 ] ),
                ( {'size':5, 'coin':5},
                  {'size':6, 'coin':6},
                  {'size':7, 'coin':7},
                  {'size':8, 'coin':8},
                  {'size':9, 'coin':9} ),
                ( 'jitter', [ 0.003] ),
                ( 'width',  [ 0.012, 0.009, 0.006, 0.003 ] ),
                ( 'target', [  'c' ] ),
                ( 'rfnids', [ [0, 4, 5] ] ),
                ( 'srng',   [ srng ] ) ] ]

    run(detect, grids, int(argv[1])-1 if len(argv) > 1 else -1)
